/**
 * @file	trianglemesh.cpp
 * @author	Andrew Robert Owens
 * @date 	2018-2-24.
 * @brief	CollisionExample: trianglemesh.cpp
 * @details Organization: Biological Modeling and Visualization research group
 * University of Calgary, Calgary, AB, Canada
 *
 * Contact:	arowens [at] ucalgary.ca
 * @copyright Copyright (c) 2018 ___ALGORITHMIC_BOTANY___. All rights reserved.
 *
 * @brief
 */

#include "triangle_mesh.h"

#include <algorithm>
#include <cassert>

#include "ordered_tuple.h"

using namespace math;

namespace geometry {

TriangleMesh::TriangleMesh(std::vector<math::Vec3f> vertices,
                           std::vector<Triangle<int>> indexTriangles)
    : m_vertices(vertices), m_indexTriangles(indexTriangles) {}

void TriangleMesh::reloadVertices(std::vector<Vec3f> vertices) {
  assert(m_vertices.size() == vertices.size());
  m_vertices = vertices;
}

std::vector<math::Vec3f> const &TriangleMesh::vertices() const {
  return m_vertices;
}

std::vector<geometry::Triangle<int>> const &TriangleMesh::triangles() const {
  return m_indexTriangles;
}

unsigned int TriangleMesh::vertexCount() const { return m_vertices.size(); }

unsigned int TriangleMesh::triangleCount() const {
  return m_indexTriangles.size();
}

geometry::Triangle<int> const *TriangleMesh::triangleData() const {
  return m_indexTriangles.data();
}

math::Vec3f const *TriangleMesh::vertexData() const {
  return m_vertices.data();
}

// FREE FUNCTTIONS
std::vector<Vec3f> calculateTriangleNormals(TriangleMesh const &mesh) {
  return calculateTriangleNormals(mesh.vertices(), mesh.triangles());
}

std::vector<Vec3f> calculateTriangleNormals(
    std::vector<math::Vec3f> const &vertices,
    std::vector<geometry::Triangle<int>> const &triangles) {
  std::vector<Vec3f> tmp;
  tmp.reserve(triangles.size());

  for (auto const &triangle : triangles) {
    auto const &a = vertices[triangle.a];
    auto const &b = vertices[triangle.b];
    auto const &c = vertices[triangle.c];

    // take cross product and normalize
    // assumes counter clockwise winding a-b-c
    tmp.emplace_back(normalized((b - a) ^ (c - a)));
  }

  return tmp;
}

std::vector<Vec3f>
calculateVertexNormals(TriangleMesh const &mesh,
                       std::vector<Vec3f> const &triangleNormals) {
  return calculateVertexNormals(mesh.vertices(), mesh.triangles(),
                                triangleNormals);
}

std::vector<Vec3f>
calculateVertexNormals(std::vector<math::Vec3f> const &vertices,
                       std::vector<geometry::Triangle<int>> const &triangles,
                       std::vector<Vec3f> const &triangleNormals) {
  assert(triangles.size() == triangleNormals.size());

  std::vector<Vec3f> tmp;
  // allocate and preinitalize to zero for following normal accumulation
  tmp.resize(vertices.size(), Vec3f(0.f, 0.f, 0.f));

  for (int tID = 0; tID < triangles.size(); ++tID) {
    auto const &t = triangles[tID];
    auto const &n = triangleNormals[tID];

    // accumulate face normal for each vertex
    tmp[t.a] += n;
    tmp[t.b] += n;
    tmp[t.c] += n;
  }

  // normalize (average) normals
  for (auto &n : tmp) {
    n.normalize();
  }

  return tmp;
}

} // namespace geometry
