/**
 * @author	Andrew Robert Owens
 * @date January, 2018
 * @brief	CPSC 587/687 Fundamental of Computer Animation
 * Organization: University of Calgary
 *
 * Contact:	arowens [at] ucalgary.ca
 *
 * Copyright (c) 2018 - Please give credit to the author.
 *
 * @file	main.cpp
 *
 * @brief
 */

// std
#include <cmath>
#include <cstdlib>
#include <iostream>
#include <limits>
#include <vector>
// graphics
#include "GLFW/glfw3.h"
#include "glad/glad.h"
// math
#include "common_matrices.h"
#include "mat4f.h"
#include "math_utility.h"
#include "ordered_tuple.h"
#include "vec3f.h"
#include "vec4f.h"
// openGL
#include "buffer_object.h"
#include "program.h"
#include "shader.h"
#include "shader_file_io.h"
#include "vertex_array_object.h"
// scene
#include "camera.h"
// geometry
#include "cylinder.h"
#include "intersection.h"
#include "ray.h"
#include "triangle.h"
#include "triangle_mesh.h"
#include "triangle_mesh_file_io.h"

//==================== GLOBAL VARIABLES ====================//
/*	Put here for simplicity. */

// Drawing Programs manager
std::vector<openGL::Program> g_program;
std::vector<openGL::VertexArrayObject> g_vao;
std::vector<openGL::BufferObject> g_buffers;

// move to another class/file
namespace simulation {
struct Object {
  math::Vec3f x;
  Object(math::Vec3f x) : x(x) {}
};

void simulateBehaviour(std::vector<Object> &objects, float deltaTime) {
  // Logic //
  /* Put here */

  // Movement //
  // stupid hack to save over multiple simulation steps
  static float accum = 0.f;
  accum += deltaTime;
  math::Vec3f offset(0.f, std::sin(M_PI * 2.f * accum), 0.f);

  for (auto &b : objects) {
    // just an arbitrary function to get the objects moving
    b.x += offset * deltaTime;
  }
}

} // namespace simulation

namespace openGL {
// Data needed rendering for mesh
struct RenderableMesh {
  GLuint vaoID = 0;
  GLuint vertexBufferID = 0;
  GLuint indexBufferID = 0;
  GLuint verticesCount = 0;
  GLuint indicesCount = 0;
  math::Mat4f modelMatrix = math::identity();
};

struct InstancedMeshes : public RenderableMesh {
  GLuint instancedBufferID = 0;
  GLuint instanceCount = 0;
};
} // namespace openGL

openGL::InstancedMeshes g_instancedMeshesData;
std::vector<simulation::Object> g_objects; // collection of things

openGL::RenderableMesh g_cylinderData;
geometry::Cylinder g_cylinder(2.f); // single cylinder

// Only one camera, so only one veiw and perspective matrix are needed.
math::Mat4f g_V;
math::Mat4f g_P;

math::Mat4f g_MVP;

// Camera and veiwing Stuff
scene::Camera g_camera;
scene::CameraUpdate g_cameraUpdate;

std::string g_modelFilePath = "susan.obj";
std::string g_cylinderFilePath = "cylinder.obj";

float g_rotationSpeed = 1.f;
float g_panningSpeed = 0.25f;
float g_cursorSpeed = 0.05f;
bool g_cursorLocked;
float g_cursorX, g_cursorY;

bool g_play = false;

int WIN_WIDTH = 1200, WIN_HEIGHT = 900;
int FB_WIDTH = WIN_WIDTH, FB_HEIGHT = WIN_HEIGHT;
float WIN_FOV = 50.f;
float WIN_NEAR = 0.01f;
float WIN_FAR = 500.f;

//==================== FUNCTION DECLARATIONS ====================//
void displayFunc();
void oncePerFrame();

void resizeFunc();

bool init();
bool generateIDs();
void deleteIDs();
void setupCylinderVAO();
void setupMeshVAO();

// load once
bool loadCylinderMeshGeometryToGPU();
bool loadMeshGeometryToGPU();

// load every frame
void reloadInstancedMatricesToGPU();

void reloadProjectionMatrix();
void reloadViewMatrix();

void windowSetSizeFunc();
void windowKeyFunc(GLFWwindow *window, int key, int scancode, int action,
                   int mods);
void windowMouseMotionFunc(GLFWwindow *window, double x, double y);
void windowSetSizeFunc(GLFWwindow *window, int width, int height);
void windowSetFramebufferSizeFunc(GLFWwindow *window, int width, int height);
void windowMouseButtonFunc(GLFWwindow *window, int button, int action,
                           int mods);
void windowMouseMotionFunc(GLFWwindow *window, double x, double y);
void windowKeyFunc(GLFWwindow *window, int key, int scancode, int action,
                   int mods);
void moveCamera();
void resetCamera();

std::string GL_ERROR();
int main(int, char **);

//==================== FUNCTION DEFINITIONS ====================//

void displayFunc() {
  using namespace openGL;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // ===== DRAW CYLINDER MESH ====== //
  // get reference to phong program
  auto &phongProgram = g_program[1];
  // Use our shader (must be called BEFORE setting uniforms!!!)
  phongProgram.use();

  setUniformMat4f(phongProgram.uniformLocation("MVP"),
                  g_P * g_V * g_cylinderData.modelMatrix,
                  true); // true, transpose for stupid OpenGL
  setUniformMat4f(phongProgram.uniformLocation("M"), g_cylinderData.modelMatrix,
                  true); // true, transpose for stupid OpenGL
  setUniformVec3f(phongProgram.uniformLocation("lightPosition_worldSpace"),
                  {100.f, 100.f, 100.f});
  setUniformVec3f(phongProgram.uniformLocation("cameraPosition_worldSpace"),
                  g_camera.localPosition());

  // Use VAO that holds buffer bindings
  // and attribute config of buffers
  g_vao[g_cylinderData.vaoID].bind();
  glDrawElements(GL_TRIANGLES,                //
                 g_cylinderData.indicesCount, //
                 GL_UNSIGNED_INT,             //
                 (void *)0);

  // ===== DRAW (INSTANCED) MESHES ====== //
  // get reference to (instanced) phong program
  auto &instancedPhongProgram = g_program[2];
  // Use our shader (must be called BEFORE setting uniforms!!!)
  instancedPhongProgram.use();

  setUniformMat4f(instancedPhongProgram.uniformLocation("V"), g_V,
                  true); // true, transpose for stupid OpenGL
  setUniformMat4f(instancedPhongProgram.uniformLocation("P"), g_P,
                  true); // true, transpose for stupid OpenGL
  setUniformVec3f(
      instancedPhongProgram.uniformLocation("lightPosition_worldSpace"),
      {100.f, 100.f, 100.f});
  setUniformVec3f(
      instancedPhongProgram.uniformLocation("cameraPosition_worldSpace"),
      g_camera.localPosition());

  // Use VAO that holds buffer bindings
  // and attribute config of buffers
  g_vao[g_instancedMeshesData.vaoID].bind();
  glDrawElementsInstanced(GL_TRIANGLES,                       //
                          g_instancedMeshesData.indicesCount, //
                          GL_UNSIGNED_INT,                    //
                          (void *)0,                          //
                          g_instancedMeshesData.instanceCount);
}

void oncePerFrame() {

  // simulate behavior
  constexpr float deltaTime = 0.001f;
  for (short iter = 0; iter < 5; ++iter)
    simulation::simulateBehaviour(g_objects, deltaTime);

  // reload new positions back to GPU
  reloadInstancedMatricesToGPU();
}

void loadObjectPositions() {
  g_instancedMeshesData.instanceCount = 2000;
  srand(0);
  for (int i = 0; i < g_instancedMeshesData.instanceCount; ++i) {
    float x = 40.f * rand() / float(RAND_MAX) - 20.f;
    float y = 40.f * rand() / float(RAND_MAX) - 20.f;
    float z = 40.f * rand() / float(RAND_MAX) - 20.f;
    g_objects.emplace_back(math::Vec3f{x, y, z});
  }
}

bool loadMeshGeometryToGPU() {

  auto isMesh = geometry::loadTriangleMeshFromFile(g_modelFilePath);
  if (!isMesh)
    return false;

  auto mesh = (*isMesh);

  g_instancedMeshesData.verticesCount = mesh.vertexCount();
  g_instancedMeshesData.indicesCount = mesh.triangleCount() * 3;

  auto const &vertices = mesh.vertices();
  auto const &triangles = mesh.triangles();

  // Pack vertex attibutes
  //		Pos		Normal		RGB
  // [ Vec3f | Vec3f | Vec3f ]
  struct Vertex {
    math::Vec3f position;
    math::Vec3f normal;
    math::Vec3f rgb;
  };

  auto vertexNormals = geometry::calculateVertexNormals(
      vertices, triangles,
      geometry::calculateTriangleNormals(vertices, triangles));

  std::vector<Vertex> vertexPack;
  vertexPack.reserve(g_instancedMeshesData.verticesCount);

  for (unsigned int id = 0; id < g_instancedMeshesData.verticesCount; ++id) {
    auto const &v = vertices[id];
    auto const &n = vertexNormals[id];

    Vertex packedVertex = {
        {v.m_x, v.m_y, v.m_z}, // pos
        {n.m_x, n.m_y, n.m_z}, // normal
        {0.8f, 0.7f, 0.1f}     // color (also changed in instanced_phong_vs.glsl
    };
    vertexPack.push_back(packedVertex);
  }

  g_buffers[g_instancedMeshesData.vertexBufferID].bind(
      openGL::BufferObject::ARRAY);
  glBufferData(GL_ARRAY_BUFFER,
               sizeof(Vertex) * vertexPack.size(), // byte size of Vec3f
               vertexPack.data(), // pointer (Vec3f*) to contents of verts
               GL_STATIC_DRAW);   // Usage pattern of GPU buffer

  g_buffers[g_instancedMeshesData.indexBufferID].bind(
      openGL::BufferObject::ELEMENT_ARRAY);
  glBufferData(GL_ELEMENT_ARRAY_BUFFER,
               sizeof(GLuint) * g_instancedMeshesData.indicesCount,
               triangles.data(), GL_STATIC_DRAW);

  return true;
}

bool loadCylinderMeshToGPU() {
  auto isMesh = geometry::loadTriangleMeshFromFile(g_cylinderFilePath);
  if (!isMesh)
    return false;

  g_cylinderData.modelMatrix =
      math::translateMatrix(g_cylinder.origin()) *
      math::rotateAboutXMatrix(90.f) *
      math::scaleMatrix(g_cylinder.radius(), 1.f, g_cylinder.radius());

  auto mesh = (*isMesh);

  g_cylinderData.verticesCount = mesh.vertexCount();
  g_cylinderData.indicesCount = mesh.triangleCount() * 3;

  auto const &vertices = mesh.vertices();
  auto const &triangles = mesh.triangles();

  // Pack vertex attibutes
  //		Pos		Normal		RGB
  // [ Vec3f | Vec3f | Vec3f ]
  struct Vertex {
    math::Vec3f position;
    math::Vec3f normal;
    math::Vec3f rgb;
  };

  auto vertexNormals = geometry::calculateVertexNormals(
      vertices, triangles,
      geometry::calculateTriangleNormals(vertices, triangles));

  std::vector<Vertex> vertexPack;
  vertexPack.reserve(g_cylinderData.verticesCount);

  for (unsigned int id = 0; id < g_cylinderData.verticesCount; ++id) {
    auto const &v = vertices[id];
    auto const &n = vertexNormals[id];

    Vertex packedVertex = {
        {v.m_x, v.m_y, v.m_z}, // position
        {n.m_x, n.m_y, n.m_z}, // normal
        {0.8f, 0.7f, 0.1f}     // color
    };
    vertexPack.push_back(packedVertex);
  }

  g_buffers[g_cylinderData.vertexBufferID].bind(openGL::BufferObject::ARRAY);
  glBufferData(GL_ARRAY_BUFFER,
               sizeof(Vertex) * vertexPack.size(), // byte size of Vec3f
               vertexPack.data(), // pointer (Vec3f*) to contents of verts
               GL_STATIC_DRAW);   // Usage pattern of GPU buffer

  g_buffers[g_cylinderData.indexBufferID].bind(
      openGL::BufferObject::ELEMENT_ARRAY);
  glBufferData(GL_ELEMENT_ARRAY_BUFFER,
               sizeof(GLuint) * g_cylinderData.indicesCount, triangles.data(),
               GL_STATIC_DRAW);

  return true;
}

void reloadInstancedMatricesToGPU() {
  using namespace math;
  // static so that we don't need to realocate a vector every frame...
  // not the best practice but useful here
  static std::vector<Mat4f::array16f> matrices;
  matrices.resize(g_instancedMeshesData.instanceCount);

  std::transform(begin(g_objects), end(g_objects), begin(matrices),
                 [&](simulation::Object const &b) {
                   auto pos = b.x;
                   // NOTE: OpenGL Expects Column major ordered matrices
                   // Since we are passing raw memerory to GPU, we need to store
                   // all the Matrices as contiguous memory (array16f)

                   // Mat4f::array16f =
                   // transposed(translateMatrix(b.x).array()); or
                   return math::Mat4f::array16f{
                       1.f,     0.f,     0.f,     0.f, //
                       0.f,     1.f,     0.f,     0.f, //
                       0.f,     0.f,     1.f,     0.f, //
                       pos.x(), pos.y(), pos.z(), 1.f  //
                   };
                 });

  g_buffers[g_instancedMeshesData.instancedBufferID].bind(
      openGL::BufferObject::ARRAY);
  glBufferData(GL_ARRAY_BUFFER, matrices.size() * sizeof(Mat4f::array16f),
               matrices.data(), GL_STREAM_DRAW);
}

void setupCylinderVAO() {
  g_vao[g_cylinderData.vaoID].bind();

  g_buffers[g_cylinderData.indexBufferID].bind(
      openGL::BufferObject::ELEMENT_ARRAY);
  g_buffers[g_cylinderData.vertexBufferID].bind(openGL::BufferObject::ARRAY);
  glEnableVertexAttribArray(0);                  // match layout # in shader
  glVertexAttribPointer(0,                       // attribute layout # above
                        3,                       // # of components (ie XYZ )
                        GL_FLOAT,                // type of components
                        GL_FALSE,                // need to be normalized?
                        sizeof(math::Vec3f) * 3, // stride
                        (void *)0                // array buffer offset
  );
  glEnableVertexAttribArray(1);                     // match layout # in shader
  glVertexAttribPointer(1,                          // attribute layout # above
                        3,                          // # of components (ie XYZ )
                        GL_FLOAT,                   // type of components
                        GL_FALSE,                   // need to be normalized?
                        sizeof(math::Vec3f) * 3,    // stride
                        (void *)sizeof(math::Vec3f) // array buffer offset
  );
  glEnableVertexAttribArray(2);                  // match layout # in shader
  glVertexAttribPointer(2,                       // attribute layout # above
                        3,                       // # of components (ie XYZ )
                        GL_FLOAT,                // type of components
                        GL_FALSE,                // need to be normalized?
                        sizeof(math::Vec3f) * 3, // stride
                        (void *)(sizeof(math::Vec3f) * 2) // array buffer offset
  );
}

void setupMeshVAO() {
  g_vao[g_instancedMeshesData.vaoID].bind();

  g_buffers[g_instancedMeshesData.indexBufferID].bind(
      openGL::BufferObject::ELEMENT_ARRAY);
  g_buffers[g_instancedMeshesData.vertexBufferID].bind(
      openGL::BufferObject::ARRAY);
  glEnableVertexAttribArray(0);                  // match layout # in shader
  glVertexAttribPointer(0,                       // attribute layout # above
                        3,                       // # of components (ie XYZ )
                        GL_FLOAT,                // type of components
                        GL_FALSE,                // need to be normalized?
                        sizeof(math::Vec3f) * 3, // stride
                        (void *)0                // array buffer offset
  );
  glEnableVertexAttribArray(1);                     // match layout # in shader
  glVertexAttribPointer(1,                          // attribute layout # above
                        3,                          // # of components (ie XYZ )
                        GL_FLOAT,                   // type of components
                        GL_FALSE,                   // need to be normalized?
                        sizeof(math::Vec3f) * 3,    // stride
                        (void *)sizeof(math::Vec3f) // array buffer offset
  );
  glEnableVertexAttribArray(2);                  // match layout # in shader
  glVertexAttribPointer(2,                       // attribute layout # above
                        3,                       // # of components (ie XYZ )
                        GL_FLOAT,                // type of components
                        GL_FALSE,                // need to be normalized?
                        sizeof(math::Vec3f) * 3, // stride
                        (void *)(sizeof(math::Vec3f) * 2) // array buffer offset
  );

  g_buffers[g_instancedMeshesData.instancedBufferID].bind(
      openGL::BufferObject::ARRAY);
  // column 0
  glEnableVertexAttribArray(3);
  glVertexAttribPointer(3,                       // attribute layout # above
                        4,                       // # of components (ie XYZ )
                        GL_FLOAT,                // type of components
                        GL_FALSE,                // need to be normalized?
                        sizeof(math::Vec4f) * 4, // stride
                        (void *)(0 * sizeof(math::Vec4f)) // array buffer offset
  );
  // column 1
  glEnableVertexAttribArray(4);
  glVertexAttribPointer(4,                       // attribute layout # above
                        4,                       // # of components (ie XYZ )
                        GL_FLOAT,                // type of components
                        GL_FALSE,                // need to be normalized?
                        sizeof(math::Vec4f) * 4, // stride
                        (void *)(1 * sizeof(math::Vec4f)) // array buffer offset
  );
  // column 2
  glEnableVertexAttribArray(5);
  glVertexAttribPointer(5,                       // attribute layout # above
                        4,                       // # of components (ie XYZ )
                        GL_FLOAT,                // type of components
                        GL_FALSE,                // need to be normalized?
                        sizeof(math::Vec4f) * 4, // stride
                        (void *)(2 * sizeof(math::Vec4f)) // array buffer offset
  );
  // column 3
  glEnableVertexAttribArray(6);
  glVertexAttribPointer(6,                       // attribute layout # above
                        4,                       // # of components (ie XYZ )
                        GL_FLOAT,                // type of components
                        GL_FALSE,                // need to be normalized?
                        sizeof(math::Vec4f) * 4, // stride
                        (void *)(3 * sizeof(math::Vec4f)) // array buffer offset
  );

  // requires opengl > 3.3
  glVertexAttribDivisor(3, 1);
  glVertexAttribDivisor(4, 1);
  glVertexAttribDivisor(5, 1);
  glVertexAttribDivisor(6, 1);

  glBindVertexArray(0); // reset to default
}

void reloadProjectionMatrix() {
  g_P = math::perspectiveProjection(WIN_FOV, // FOV
                                    static_cast<float>(WIN_WIDTH) /
                                        WIN_HEIGHT, // Aspect
                                    WIN_NEAR,       // near plane
                                    WIN_FAR);       // far plane depth
}

void reloadViewMatrix() { g_V = scene::makeViewMatrix(g_camera); }

bool generateIDs() {
  using namespace openGL;
  // shader ID from OpenGL
  {
    auto vsSource = loadShaderStringFromFile("basic_vs.glsl");
    auto fsSource = loadShaderStringFromFile("basic_fs.glsl");
    if (vsSource.empty() || fsSource.empty()) {
      std::cerr << "Failed to load shaders from file\n";
      return false;
    }
    auto program = makeProgram(vsSource, fsSource);
    if (!program.isValid()) {
      std::cerr << "Failed to load program\n";
      return false;
    }
    g_program.push_back(std::move(program));
  }
  {
    auto vsSource = loadShaderStringFromFile("phong_vs.glsl");
    auto fsSource = loadShaderStringFromFile("phong_fs.glsl");
    if (vsSource.empty() || fsSource.empty()) {
      std::cerr << "Failed to load shaders from file\n";
      return false;
    }
    auto program = makeProgram(vsSource, fsSource);
    if (!program.isValid()) {
      std::cerr << "Failed to load program\n";
      return false;
    }
    g_program.push_back(std::move(program));
  }
  {
    auto vsSource = loadShaderStringFromFile("instanced_phong_vs.glsl");
    auto fsSource = loadShaderStringFromFile("phong_fs.glsl");
    if (vsSource.empty() || fsSource.empty()) {
      std::cerr << "Failed to load shaders from file\n";
      return false;
    }
    auto program = makeProgram(vsSource, fsSource);
    if (!program.isValid()) {
      std::cerr << "Failed to load program\n";
      return false;
    }
    g_program.push_back(std::move(program));
  }

  // VAO and buffer IDs given from OpenGL
  // Cylinder
  {
    auto vao = makeVertexArrayObject();
    if (!vao.isValid()) {
      std::cerr << "Failed to load vao\n";
      return false;
    }
    g_cylinderData.vaoID = g_vao.size();
    g_vao.push_back(std::move(vao));

    auto vertexBuffer = makeBufferObject();
    if (!vertexBuffer.isAllocated()) {
      std::cerr << "Failed to make buffer\n";
      return false;
    }
    g_cylinderData.vertexBufferID = g_buffers.size();
    g_buffers.push_back(std::move(vertexBuffer));

    auto indexBuffer = makeBufferObject();
    if (!indexBuffer.isAllocated()) {
      std::cerr << "Failed to make buffer\n";
      return false;
    }
    g_cylinderData.indexBufferID = g_buffers.size();
    g_buffers.push_back(std::move(indexBuffer));
  }

  // Boids
  {
    auto vao = makeVertexArrayObject();
    if (!vao.isValid()) {
      std::cerr << "Failed to load vao\n";
      return false;
    }
    g_instancedMeshesData.vaoID = g_vao.size();
    g_vao.push_back(std::move(vao));

    auto vertexBuffer = makeBufferObject();
    if (!vertexBuffer.isAllocated()) {
      std::cerr << "Failed to make buffer\n";
      return false;
    }
    g_instancedMeshesData.vertexBufferID = g_buffers.size();
    g_buffers.push_back(std::move(vertexBuffer));

    auto indexBuffer = makeBufferObject();
    if (!indexBuffer.isAllocated()) {
      std::cerr << "Failed to make buffer\n";
      return false;
    }
    g_instancedMeshesData.indexBufferID = g_buffers.size();
    g_buffers.push_back(std::move(indexBuffer));

    auto transformsBuffer = makeBufferObject();
    if (!transformsBuffer.isAllocated()) {
      std::cerr << "Failed to make buffer\n";
      return false;
    }
    g_instancedMeshesData.instancedBufferID = g_buffers.size();
    g_buffers.push_back(std::move(transformsBuffer));
  }

  return true;
}

void deleteIDs() {
  g_program.clear(); // calls destructors on shaders, deallocates GPU
  g_vao.clear();
  g_buffers.clear();
}

bool init() {
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  // glEnable(GL_MULTISAMPLE);
  glPointSize(50);

  if (!generateIDs())
    return false;

  setupMeshVAO();
  setupCylinderVAO();

  if (!loadMeshGeometryToGPU())
    return false;

  if (!loadCylinderMeshToGPU())
    return false;

  loadObjectPositions();
  reloadInstancedMatricesToGPU();

  resetCamera();

  reloadProjectionMatrix();
  reloadViewMatrix();

  return true;
}

int main(int argc, char **argv) {

  GLFWwindow *window;

  if (!glfwInit()) {
    exit(EXIT_FAILURE);
  }

  int glfwMajor, glfwMinor, glfwRevision;
  glfwGetVersion(&glfwMajor, &glfwMinor, &glfwRevision);
  std::cout << "GLFW Version: " << glfwMajor << "." << glfwMinor << "."
            << glfwRevision << std::endl;

  glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);
  glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
  glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
  glfwWindowHint(GLFW_SAMPLES, 4);

  const GLFWvidmode *mode = glfwGetVideoMode(glfwGetPrimaryMonitor());
  glfwWindowHint(GLFW_RED_BITS, mode->redBits);
  glfwWindowHint(GLFW_GREEN_BITS, mode->greenBits);
  glfwWindowHint(GLFW_BLUE_BITS, mode->blueBits);
  glfwWindowHint(GLFW_REFRESH_RATE, mode->refreshRate);

  window = glfwCreateWindow(WIN_WIDTH, WIN_HEIGHT, "Boids", NULL, NULL);
  if (!window) {
    glfwTerminate();
    exit(EXIT_FAILURE);
  }

  glfwGetWindowSize(window, &WIN_WIDTH, &WIN_HEIGHT);
  glfwGetFramebufferSize(window, &FB_WIDTH, &FB_HEIGHT);

  glfwSetWindowSizeCallback(window, windowSetSizeFunc);
  glfwSetFramebufferSizeCallback(window, windowSetFramebufferSizeFunc);
  glfwSetKeyCallback(window, windowKeyFunc);
  glfwSetCursorPosCallback(window, windowMouseMotionFunc);
  glfwSetMouseButtonCallback(window, windowMouseButtonFunc);

  glfwMakeContextCurrent(window);
  gladLoadGLLoader((GLADloadproc)glfwGetProcAddress);
  glfwSwapInterval(0);

  std::cout << "GL Version: :" << glGetString(GL_VERSION) << std::endl;
  std::cout << GL_ERROR() << std::endl;

  if (init()) // our own initialize stuff func
  {
    while (!glfwWindowShouldClose(window)) {
      moveCamera();
      if (g_play) {
        oncePerFrame();
      }
      displayFunc();

      glfwSwapBuffers(window);
      glfwPollEvents();
    }
  }

  // clean up after loop
  deleteIDs();

  glfwDestroyWindow(window);
  glfwTerminate();

  return EXIT_SUCCESS;
}

//==================== CALLBACK FUNCTIONS ====================//

void windowSetSizeFunc(GLFWwindow *window, int width, int height) {
  WIN_WIDTH = width;
  WIN_HEIGHT = height;

  reloadProjectionMatrix();
}

void windowSetFramebufferSizeFunc(GLFWwindow *window, int width, int height) {
  FB_WIDTH = width;
  FB_HEIGHT = height;

  glViewport(0, 0, FB_WIDTH, FB_HEIGHT);
}

void windowMouseButtonFunc(GLFWwindow *window, int button, int action,
                           int mods) {
  if (button == GLFW_MOUSE_BUTTON_LEFT) {
    if (action == GLFW_PRESS) {
      g_cursorLocked = GL_TRUE;
    } else {
      g_cursorLocked = GL_FALSE;
    }
  }
}

void windowMouseMotionFunc(GLFWwindow *window, double x, double y) {
  if (g_cursorLocked) {
    float deltaX = (x - g_cursorX) * g_cursorSpeed;
    float deltaY = (y - g_cursorY) * g_cursorSpeed;
    g_camera.rotateRightAroundFocus(deltaX);
    g_camera.rotateDownAroundFocus(deltaY);

    reloadViewMatrix();
  }

  g_cursorX = x;
  g_cursorY = y;
}

void windowKeyFunc(GLFWwindow *window, int key, int scancode, int action,
                   int mods) {
  using namespace scene;

  if (action != GLFW_REPEAT) {
    bool set = (action != GLFW_RELEASE);
    switch (key) {
    case GLFW_KEY_ESCAPE:
      glfwSetWindowShouldClose(window, GL_TRUE);
      break;
    case GLFW_KEY_A:
      g_cameraUpdate.set(CameraUpdate::moveLeft, set);
      break;
    case GLFW_KEY_D:
      g_cameraUpdate.set(CameraUpdate::moveRight, set);
      break;
    case GLFW_KEY_E:
      g_cameraUpdate.set(CameraUpdate::moveUp, set);
      break;
    case GLFW_KEY_Q:
      g_cameraUpdate.set(CameraUpdate::moveDown, set);
      break;
    case GLFW_KEY_R:
      if (mods == GLFW_MOD_CONTROL)
        g_play = true;
      break;
    case GLFW_KEY_S:
      if (mods == GLFW_MOD_CONTROL)
        g_play = false;
      else
        g_cameraUpdate.set(CameraUpdate::moveBackward, set);
      break;
    case GLFW_KEY_V:
      resetCamera();
      break;
    case GLFW_KEY_W:
      g_cameraUpdate.set(CameraUpdate::moveForward, set);
      break;
    case GLFW_KEY_UP:
      g_cameraUpdate.set(CameraUpdate::rotateUp, set);
      break;
    case GLFW_KEY_DOWN:
      g_cameraUpdate.set(CameraUpdate::rotateDown, set);
      break;
    case GLFW_KEY_LEFT:
      if (mods == GLFW_MOD_SHIFT)
        g_cameraUpdate.set(CameraUpdate::rollLeft, set);
      else
        g_cameraUpdate.set(CameraUpdate::rotateLeft, set);
      break;
    case GLFW_KEY_RIGHT:
      if (mods == GLFW_MOD_SHIFT)
        g_cameraUpdate.set(CameraUpdate::rollRight, set);
      else
        g_cameraUpdate.set(CameraUpdate::rotateRight, set);
      break;
    case GLFW_KEY_SPACE:
      g_play = set ? !g_play : g_play;
      break;
    case GLFW_KEY_LEFT_BRACKET:
      if (mods == GLFW_MOD_SHIFT) {
        g_rotationSpeed *= 0.5f;
      } else {
        g_panningSpeed *= 0.5f;
      }
      break;
    case GLFW_KEY_RIGHT_BRACKET:
      if (mods == GLFW_MOD_SHIFT) {
        g_rotationSpeed *= 1.5f;
      } else {
        g_panningSpeed *= 1.5f;
      }
      break;
    default:
      break;
    }
  }
}

//==================== OPENGL HELPER FUNCTIONS ====================//

void moveCamera() {
  using namespace scene;
  if (g_cameraUpdate.needsUpdating()) {
    if (g_cameraUpdate.isSet(CameraUpdate::moveBackward)) {
      g_camera.moveBackward(g_panningSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::moveForward)) {
      g_camera.moveForward(g_panningSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::moveUp)) {
      g_camera.moveUp(g_panningSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::moveDown)) {
      g_camera.moveDown(g_panningSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::moveLeft)) {
      g_camera.moveLeft(g_panningSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::moveRight)) {
      g_camera.moveRight(g_panningSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::rotateLeft)) {
      g_camera.rotateLeft(g_rotationSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::rotateRight)) {
      g_camera.rotateRight(g_rotationSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::rotateUp)) {
      g_camera.rotateUp(g_rotationSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::rotateDown)) {
      g_camera.rotateDown(g_rotationSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::rollLeft)) {
      g_camera.rollLeft(g_rotationSpeed);
    }
    if (g_cameraUpdate.isSet(CameraUpdate::rollRight)) {
      g_camera.rollRight(g_rotationSpeed);
    }

    reloadViewMatrix();
    // g_cameraUpdate.reset(); // reseting seems jittery, so don't
  }
}

void resetCamera() {
  g_camera = scene::Camera(math::Vec3f{0.f, 0.f, 5.f},  // position
                           math::Vec3f{0.f, 0.f, -1.f}, // forward
                           math::Vec3f{0.f, 1.f, 0.f}   // up
  );
}

std::string GL_ERROR() {
  GLenum code = glGetError();

  switch (code) {
  case GL_NO_ERROR:
    return "GL_NO_ERROR";
  case GL_INVALID_ENUM:
    return "GL_INVALID_ENUM";
  case GL_INVALID_VALUE:
    return "GL_INVALID_VALUE";
  case GL_INVALID_OPERATION:
    return "GL_INVALID_OPERATION";
  case GL_INVALID_FRAMEBUFFER_OPERATION:
    return "GL_INVALID_FRAMEBUFFER_OPERATION";
  case GL_OUT_OF_MEMORY:
    return "GL_OUT_OF_MEMORY";
  default:
    return "Non Valid Error Code";
  }
}
