//
//  math_utility.tpp
//  CollisionExample
//
//  Created by Andrew Owens on 2018-02-21.
//  Copyright (c) 2018 ___ALGORITHMIC_BOTANY___. All rights reserved.
//

#include <algorithm>
#include <cmath>
#include <functional>
#include <limits>

namespace math {

template <typename Float> constexpr Float radianToDegree(Float radian) {
  static_assert(std::is_floating_point<Float>::value, "");

  return Float(180) / Float(M_PI) * radian;
}

template <typename Float> constexpr Float degreeToRadian(Float degree) {
  static_assert(std::is_floating_point<Float>::value, "");

  return Float(M_PI) / Float(180) * degree;
}

template <typename Float> constexpr bool isZero(Float x) {
  static_assert(std::is_floating_point<Float>::value, "");

  using std::abs;

  return abs(x) < std::numeric_limits<Float>::epsilon();
}

template <typename Float> constexpr bool isZero(Float x, Float epsilon) {
  static_assert(std::is_floating_point<Float>::value, "");
  using std::abs;

  return abs(x) < epsilon;
}

template <typename Signed> constexpr int sgn(Signed val) {
  static_assert(std::is_signed<Signed>::value, "");

  return (Signed(0) < val) - (val < Signed(0));
}

template <typename Signed> constexpr Signed abs(const Signed &value) {
  static_assert(std::is_signed<Signed>::value, "");
  return sgn(value) * value;
}

template <typename Type>
constexpr Type const &minimum(Type const &a, Type const &b) {
  return b < a ? b : a;
}

template <typename Type, typename Comp>
constexpr Type const &minimum(Type const &a, Type const &b, Comp comp) {
  return comp(b, a) ? b : a;
}

template <typename T, typename... Args>
constexpr T const &minimum(T const &a, T const &b, Args... args) {
  return minimum(minimum(a, b), args...);
}

template <typename T, typename... Args, typename Comp>
constexpr T const &minimum(T const &a, T const &b, Args... args, Comp comp) {
  return minimum(minimum(a, b, comp), args..., comp);
}

template <typename Type>
constexpr Type const &maximum(Type const &a, Type const &b) {
  return a > b ? a : b;
}

template <typename Type, typename Comp>
constexpr Type const &maximum(Type const &a, Type const &b, Comp comp) {
  return comp(a, b) ? a : b;
}

template <typename T, typename... Args>
constexpr T const &maximum(T const &a, T const &b, Args... args) {
  return maximum(maximum(a, b), args...);
}

template <typename T, typename... Args, typename Comp>
constexpr T const &maximum(T const &a, T const &b, Args... args, Comp comp) {
  return maximum(maximum(a, b, comp), args..., comp);
}

template <class T>
constexpr T const &clamp(const T &v, const T &lo, const T &hi) {
  // assert(!(hi < lo)); not in gcc constexpr & assert
  return minimum(maximum(lo, v), hi);
}

template <class T, class Compare>
constexpr T const &clamp(const T &v, const T &lo, const T &hi, Compare comp) {
  // using a strict comparitor comp, min(max(lo, v), hi) does not work
  return comp(v, lo) ? lo : comp(hi, v) ? hi : v;
}

} // namespace math
