/**
 * @file	buffer_object.cpp
 * @author	Andrew Robert Owens
 * @date 	2018-2-26.
 * @brief	CollisionExample: buffer_object.cpp
 * @details Organization: Biological Modeling and Visualization research group
 * University of Calgary, Calgary, AB, Canada
 *
 * Contact:	arowens [at] ucalgary.ca
 * @copyright Copyright (c) 2018 ___ALGORITHMIC_BOTANY___. All rights reserved.
 *
 * @brief
 */

#include "buffer_object.h"

#include <cassert>
#include <utility>

namespace openGL {

BufferObject::BufferObject(GLuint id) : m_id(id) {}

BufferObject::~BufferObject() { release(); }

BufferObject::BufferObject(BufferObject &&other)
    : m_id(other.m_id), m_type(other.m_type) {
  other.m_id = INVALID_ID;
  other.m_type = INVALID;
}

BufferObject &BufferObject::operator=(BufferObject &&other) {
  if (this != &other) {
    release();
    swap(*this, other);
  }

  return *this;
}

BufferObject::Type BufferObject::type() const { return m_type; }

bool BufferObject::isAllocated() const { return isValidBufferObjectID(m_id); }

bool BufferObject::isValidType() const {
  return isValidBufferObjectType(m_type);
}

bool BufferObject::isValid() const { return isAllocated() && isValidType(); }

void BufferObject::bind() const {
  assert(isValid());
  glBindBuffer(m_type, m_id);
}

void BufferObject::bind(BufferObject::Type type) {
  assert(isAllocated());
  assert(isValidBufferObjectType(enumToBufferObjectType(type)));
  glBindBuffer(type, m_id);
  m_type = type;
}

void BufferObject::release() {
  glDeleteBuffers(1, &m_id);
  m_id = BufferObject::INVALID_ID;
  m_type = BufferObject::INVALID;
}

/** FREE FUNCTIONS **/

BufferObject makeBufferObject() {
  GLuint id = BufferObject::INVALID_ID;

  glGenBuffers(1, &id);

  if (!isValidBufferObjectID(id))
    return BufferObject(BufferObject::INVALID_ID);

  return BufferObject(id);
}

void swap(BufferObject &lhs, BufferObject &rhs) {
  std::swap(lhs.m_id, rhs.m_id);
  std::swap(lhs.m_type, rhs.m_type);
}

bool isValidBufferObjectID(GLuint id) { return id != BufferObject::INVALID_ID; }

bool isValidBufferObjectType(BufferObject::Type type) {
  return type != BufferObject::INVALID;
}

BufferObject::Type enumToBufferObjectType(GLenum bufferObjectEnum) {
  switch (bufferObjectEnum) {
  case BufferObject::ARRAY:
  case BufferObject::ELEMENT_ARRAY:
  case BufferObject::TEXTURE:
    return static_cast<BufferObject::Type>(bufferObjectEnum);
  default:
    return BufferObject::INVALID;
  }
}

} // namespace openGL
