#include "program.h"

#include <iostream>
#include <vector>

namespace openGL {

Program::Program(GLuint programID) : m_id(programID) {}

Program::~Program() { release(); }

Program::Program(Program &&other) : m_id(other.m_id) { other.m_id = 0; }

Program &Program::operator=(Program &&other) {
  if (this != &other) {
    release();
    std::swap(m_id, other.m_id);
  }
  return *this;
}

void Program::use() const { glUseProgram(m_id); }

bool Program::isValid() const { return m_id != Program::INVALID_ID; }

GLint Program::uniformLocation(const std::string &name) {
  return glGetUniformLocation(m_id, name.c_str());
}

GLint Program::uniformLocation(GLchar const *name) {
  return glGetUniformLocation(m_id, name);
}

void Program::release() {
  glDeleteProgram(m_id);
  m_id = 0;
}

Program makeProgram(std::string const &vertexShaderSource,
                    std::string const &fragmentShaderSource) {
  auto vs = makeShader(vertexShaderSource, Shader::VERTEX);
  if (!vs.isValid()) {
    std::cerr << "Failed to compile vertex shader\n";
    return Program(Program::INVALID_ID);
  }
  auto fs = makeShader(fragmentShaderSource, Shader::FRAGMENT);
  if (!fs.isValid()) {
    std::cerr << "Failed to compile fragment shader\n";
    return Program(Program::INVALID_ID);
  }
  return makeProgram(vs, fs);
}

Program makeProgram(Shader const &vertexShader, Shader const &fragmentShader) {
  if (!vertexShader.isValid() || !fragmentShader.isValid()) {
    return Program(Program::INVALID_ID);
  }

  if ((vertexShader.type() != Shader::VERTEX) ||
      (fragmentShader.type() != Shader::FRAGMENT)) {
    return Program(Program::INVALID_ID);
  }

  GLuint id = glCreateProgram();
  if (!isValidProgramID(id)) {
    return Program(Program::INVALID_ID);
  }

  glAttachShader(id, vertexShader.id());
  glAttachShader(id, fragmentShader.id());

  glLinkProgram(id);

  // detach so shaders may be deallocated properly
  glDetachShader(id, vertexShader.id());
  glDetachShader(id, fragmentShader.id());

  if (!checkLinkStatus(id)) {
    glDeleteProgram(id);
    return Program(Program::INVALID_ID);
  }

  return Program(id);
}

Program makeProgram(Shader const &vertexShader, Shader const &geometryShader,
                    Shader const &fragmentShader) {
  if (!vertexShader.isValid() || !fragmentShader.isValid() ||
      !geometryShader.isValid()) {
    return Program(Program::INVALID_ID);
  }

  if ((vertexShader.type() != Shader::VERTEX) ||
      (geometryShader.type() != Shader::GEOMETRY) ||
      (fragmentShader.type() != Shader::FRAGMENT)) {
    return Program(Program::INVALID_ID);
  }

  GLuint id = glCreateProgram();
  if (!isValidProgramID(id)) {
    return Program(Program::INVALID_ID);
  }

  glAttachShader(id, vertexShader.id());
  glAttachShader(id, geometryShader.id());
  glAttachShader(id, fragmentShader.id());

  glLinkProgram(id);

  glDetachShader(id, vertexShader.id());
  glDetachShader(id, geometryShader.id());
  glDetachShader(id, fragmentShader.id());

  if (!checkLinkStatus(id)) {
    glDeleteProgram(id);
    return Program(Program::INVALID_ID);
  }

  return Program(id);
}

bool isValidProgramID(GLuint id) { return id != Program::INVALID_ID; }

bool checkLinkStatus(GLuint programID) {
  GLint result;
  int infoLogLength;
  glGetProgramiv(programID, GL_LINK_STATUS, &result);
  if (!result) {
    glGetProgramiv(programID, GL_INFO_LOG_LENGTH, &infoLogLength);
    std::vector<char> errorMsg(infoLogLength);
    glGetProgramInfoLog(programID, infoLogLength, nullptr, errorMsg.data());
    std::cerr << errorMsg.data() << std::endl;
    return false;
  }

  return true;
}

void setUniformVec3f(GLuint uniformLocation, float x, float y, float z) {
  glUniform3f(uniformLocation, x, y, z);
}

void setUniformVec3f(GLuint uniformLocation, math::Vec3f const &vec) {
  glUniform3fv(uniformLocation, 1, vec.data());
}

void setUniformVec3f(GLuint uniformLocation, GLuint count,
                     float const *vecPtr) {
  glUniform3fv(uniformLocation, count, vecPtr);
}

void setUniformMat4f(GLuint uniformLocation, math::Mat4f const &value,
                     GLboolean applyTranspose) {
  glUniformMatrix4fv(uniformLocation, 1, applyTranspose, value.data());
}

void setUniform1f(GLuint uniformLocation, float value) {
  glUniform1f(uniformLocation, value);
}

} // namespace openGL
